/**
 * Copyright (c) 2010-2012 EBM WebSourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.log.formatter;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Deque;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.Locale;
import java.util.Map;
import java.util.logging.Formatter;
import java.util.logging.Handler;
import java.util.logging.LogManager;
import java.util.logging.LogRecord;

import org.ow2.petals.commons.log.LogData;
import org.ow2.petals.commons.log.PetalsExecutionContext;

import com.ebmwebsourcing.easycommons.thread.ExecutionContext;

/**
 * This formatter is aimed at using first {@link LogRecord} parameter as "just in time evaluated" extra log information.
 * <br/>
 * Note that localization is still possible (parameters are shifted left if necessary).<br/>
 * <br/>
 * In other words :
 * <ul>
 * <li>if first parameter is an instance of {@link LogData}, this data will be appended in log message,</li>
 * <li>otherwise, first parameter is kept for localization duty, as usual.</li>.
 * 
 * Important: this Formatter is NOT thread-safe, use one instance per {@link Handler} (which are meant to be
 * thread-safe).
 * 
 * @author mjambert
 * 
 */
public class LogDataFormatter extends Formatter {

    private static final String LOGDATAFORMATTER_PROPERTY_NAME = LogDataFormatter.class.getName();

    private static final String LOGDATAFORMATTER_APPENDER_PROPERTY_NAME = LOGDATAFORMATTER_PROPERTY_NAME
            + ".appender";

    private static final String LOGDATAFORMATTER_STARTING_DELIMITER_PROPERTY_NAME = LOGDATAFORMATTER_PROPERTY_NAME
            + ".starting-delimiter";

    private static final String LOGDATAFORMATTER_DEFAULT_STARTING_DELIMITER = "";

    private static final String LOGDATAFORMATTER_ENDING_DELIMITER_PROPERTY_NAME = LOGDATAFORMATTER_PROPERTY_NAME
            + ".ending-delimiter";

    private static final String LOGDATAFORMATTER_DEFAULT_ENDING_DELIMITER = "";

    // careful, DateFormat is not thread safe!
    public final DateFormat DATE_FORMAT = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss,SSS 'GMT'Z", Locale.getDefault());

    private final Deque<LogDataAppender> dataAppenders;

    private String startingDelimiter;

    private String endingDelimiter;

    public LogDataFormatter() {
        final LogManager manager = LogManager.getLogManager();

        // Appender configuration
        {
            final String appender = manager.getProperty(LOGDATAFORMATTER_APPENDER_PROPERTY_NAME);
            LogDataAppender logDataAppender;
            if (appender != null && !appender.trim().isEmpty()) {
                try {
                    final Class<?> appenderClz = ClassLoader.getSystemClassLoader().loadClass(
                            appender);
                    final Object appenderObj = appenderClz.newInstance();
                    if (appenderObj instanceof LogDataAppender) {
                        logDataAppender = (LogDataAppender) appenderObj;
                    } else {
                        logDataAppender = new BasicLogDataAppender();
                    }
                } catch (final ClassNotFoundException e) {
                    logDataAppender = new BasicLogDataAppender();
                } catch (final InstantiationException e) {
                    logDataAppender = new BasicLogDataAppender();
                } catch (final IllegalAccessException e) {
                    logDataAppender = new BasicLogDataAppender();
                }
            } else {
                logDataAppender = new BasicLogDataAppender();
            }

            this.dataAppenders = new LinkedList<LogDataAppender>();
            this.addLogDataAppender(logDataAppender);
        }

        // Starting delimiter configuration
        {
            final String tmpStartingDelimiter = manager
                    .getProperty(LOGDATAFORMATTER_STARTING_DELIMITER_PROPERTY_NAME);
            if (tmpStartingDelimiter != null && !tmpStartingDelimiter.trim().isEmpty()) {
                this.startingDelimiter = tmpStartingDelimiter;
            } else {
                this.startingDelimiter = LOGDATAFORMATTER_DEFAULT_STARTING_DELIMITER;
            }
        }

        // Ending delimiter configuration
        {
            final String tmpEndingDelimiter = manager
                    .getProperty(LOGDATAFORMATTER_ENDING_DELIMITER_PROPERTY_NAME);
            if (tmpEndingDelimiter != null && !tmpEndingDelimiter.trim().isEmpty()) {
                this.endingDelimiter = tmpEndingDelimiter;
            } else {
                this.endingDelimiter = LOGDATAFORMATTER_DEFAULT_ENDING_DELIMITER;
            }
        }
    }

    public void setExtraParametersDelimiter(final String startingDelimiter,
            final String endingDelimiter) {
        this.startingDelimiter = startingDelimiter;
        this.endingDelimiter = endingDelimiter;
    }

    public final void addLogDataAppender(LogDataAppender logDataAppender) {
        assert logDataAppender != null;
        dataAppenders.add(logDataAppender);
    }

    final LogDataAppender[] getLogDataAppenders() {
        return dataAppenders.toArray(new LogDataAppender[dataAppenders.size()]);
    }

    protected void formatExtraParameters(StringBuilder outBuffer, Map<String, Object> extraParameters) {
        Iterator<String> extraParametersKeySetIterator = extraParameters.keySet().iterator();
        if (extraParametersKeySetIterator.hasNext() && !dataAppenders.isEmpty()) {
            outBuffer.append(this.startingDelimiter);
            boolean isFirstAppending = true;
            while (extraParametersKeySetIterator.hasNext()) {
                String extraParameterKey = extraParametersKeySetIterator.next();
                Iterator<LogDataAppender> itAppender = dataAppenders.descendingIterator();
                while (itAppender.hasNext()) {
                    LogDataAppender appender = (LogDataAppender) itAppender.next();
                    if (!appender.canAppend(extraParameters, extraParameterKey)) {
                        continue;
                    }
                    if (isFirstAppending) {
                        isFirstAppending = false;
                    } else {
                        outBuffer.append(", ");
                    }
                    appender.append(outBuffer, extraParameters, extraParameterKey);
                    break;
                }
            }
            outBuffer.append(this.endingDelimiter);
        }
    }

    private boolean isThereLogDataParameter(Object[] logParameters) {
        return (logParameters != null)
                && (logParameters.length != 0)
                && logParameters[0] instanceof LogData;
    }

    protected void populateExtraParameters(final Map<String, Object> extraParameters,
            LogRecord logRecord) {
        Object[] logParameters = logRecord.getParameters();
        if (isThereLogDataParameter(logParameters)) {
            extraParameters.putAll((Map<String, Object>) logParameters[0]);
        }
    }

    protected StringBuilder buildPattern(LogRecord record, final Map<String, Object> flowParameters) {
        StringBuilder sb = new StringBuilder();
        String formattedLog;
        final String prefix = ExecutionContext.getProperties().getProperty(
                PetalsExecutionContext.CONTAINER_NAME_PROPERTY_NAME);
        if (prefix == null) {
        	formattedLog = String.format("%s %s [%s] : %s", DATE_FORMAT.format(record.getMillis()),
                record.getLevel(), record.getLoggerName(), record.getMessage());
        } else {
        	formattedLog = String.format("%s %s %s [%s] : %s", prefix, DATE_FORMAT.format(record.getMillis()),
                    record.getLevel(), record.getLoggerName(), record.getMessage());
        }
        sb.append(formattedLog);
        if (flowParameters.size() > 0 && record.getMessage() != null
                && !record.getMessage().isEmpty()) {
            sb.append(' ');
        }
        return sb;
    }

    @Override
    public String format(LogRecord record) {
        final Map<String, Object> extraParameters = new LinkedHashMap<String, Object>();
        populateExtraParameters(extraParameters, record);

        StringBuilder sb = buildPattern(record, extraParameters);
        formatExtraParameters(sb, extraParameters);
        sb.append("\n");

        if (record.getThrown() != null) {
            final StringWriter sw = new StringWriter();
            record.getThrown().printStackTrace(new PrintWriter(sw));
            sw.flush();
            sb.append(sw.toString());
        }
        return sb.toString();
    }

    @Override
    public synchronized String formatMessage(LogRecord record) {
        Object[] extraParameters = record.getParameters();
        String result = null;
        if (isThereLogDataParameter(extraParameters)) {
            Object[] extraParametersWithoutLogData = Arrays.copyOfRange(extraParameters, 1,
                    extraParameters.length);
            record.setParameters(extraParametersWithoutLogData);
            result = super.formatMessage(record);
            record.setParameters(extraParameters);
        } else {
            result = super.formatMessage(record);
        }
        return result;
    }
}
