/**
 * Copyright (c) 2011-2012 EBM WebSourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.log.handler;

import static org.ow2.petals.commons.log.PetalsExecutionContext.FLOW_INSTANCE_ID_PROPERTY_NAME;

import java.io.File;
import java.io.IOException;
import java.util.Properties;
import java.util.logging.ErrorManager;
import java.util.logging.LogRecord;

import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.Source;

import org.apache.commons.pool.impl.GenericObjectPool;
import org.ow2.petals.commons.log.FlowLogData;
import org.ow2.petals.commons.log.Level;
import org.ow2.petals.commons.log.LogData;
import org.ow2.petals.commons.log.PetalsExecutionContext;
import org.ow2.petals.commons.log.TraceCode;

import com.ebmwebsourcing.easycommons.io.FileSystemHelper;
import com.ebmwebsourcing.easycommons.properties.PropertiesException;
import com.ebmwebsourcing.easycommons.properties.PropertiesHelper;
import com.ebmwebsourcing.easycommons.thread.ExecutionContext;

/**
 * <p>
 * A contextual file handler based on {@link PetalsFileHandler} and dumping
 * message content.
 * </p>
 */
public class PetalsPayloadDumperFileHandler extends PetalsFileHandler {

    private static final String DUMP_FILE_EXTENSION = ".xml";

    private static final String FILENAME_SEPARATOR = "_";

    public static final String PAYLOAD_CONTENT_DUMP_FILE_LOGDATA_NAME = "payloadContentDumpFile";

    protected static final String DUMP_BASEDIR_PROPERTY_NAME = ".dump-basedir";

    protected static final String TRANSFORMERS_POOL_MAX_ACTIVE_PROPERTY_NAME = ".transformersPoolMaxActive";

    protected static final String TRANSFORMERS_POOL_MAX_IDLE_PROPERTY_NAME = ".transformersPoolMaxIdle";

    private static final String JBI_OUT_MESSAGE = "out";

    private static final String JBI_IN_MESSAGE = "in";

    /**
     * The root directory (absolute) of sub directory of dumps.
     * 
     * If the value of the configuration property {@value #DUMP_BASEDIR_PROPERTY_NAME} is not an absolute path, the dump
     * dir is defined as relative to {@link #flowsSubdir} .
     */
    protected final File dumpBasedir;

    private final SourceHelpers helpers;

    public PetalsPayloadDumperFileHandler() throws PropertiesException {
        super();

        final int transformerPoolMaxActive = getIntProperty(TRANSFORMERS_POOL_MAX_ACTIVE_PROPERTY_NAME,
                GenericObjectPool.DEFAULT_MAX_ACTIVE);

        final int transformerPoolMaxIdle = getIntProperty(TRANSFORMERS_POOL_MAX_IDLE_PROPERTY_NAME,
                GenericObjectPool.DEFAULT_MAX_IDLE);

        this.helpers = new SourceHelpers(transformerPoolMaxActive, transformerPoolMaxIdle);

        // Initialize the dump base dir
        final String dumpBaseDirValue = getStringProperty(DUMP_BASEDIR_PROPERTY_NAME,
                this.flowsSubdir.getAbsolutePath());

        final File dumpBasedirPath = new File(dumpBaseDirValue);
        if (dumpBasedirPath.isAbsolute()) {
            this.dumpBasedir = dumpBasedirPath;
        } else {
            this.dumpBasedir = new File(this.flowsSubdir, dumpBaseDirValue);
        }
    }

    @Override
    public synchronized void publish(final LogRecord record) {

        int levelValue = getLevel().intValue();
        if (record.getLevel().intValue() >= levelValue && levelValue != offValue) {

            this.dumpMessageExchange(record);

            this.printLog(record);
        }
    }

    private void dumpMessageExchange(final LogRecord record) {

        final Object[] logParameters = record.getParameters();
        if (Level.MONIT.equals(record.getLevel()) && logParameters != null && logParameters.length != 0
                && logParameters[0] != null && logParameters[0] instanceof LogData) {

            final LogData logData = (LogData) logParameters[0];
            final String flowInstanceId = (String) logData
                    .get(PetalsExecutionContext.FLOW_INSTANCE_ID_PROPERTY_NAME);
            final String flowStepId = (String) logData
                    .get(PetalsExecutionContext.FLOW_STEP_ID_PROPERTY_NAME);

            assert flowInstanceId != null : "A MONIT log event MUST always have an attribute 'flowInstanceId' set !!!";
            assert !flowInstanceId.trim().isEmpty() : "A MONIT log event MUST always have a non-empty attribute 'flowInstanceId' !!!";
            assert flowStepId != null : "A MONIT log event MUST always have an attribute 'flowStepId' set !!!";
            assert !flowStepId.trim().isEmpty() : "A MONIT log event MUST always have a non-empty attribute 'flowStepId' !!!";

            final MessageExchange exchange = (MessageExchange) logData
                    .get(PetalsExecutionContext.FLOW_EXCHANGE_PROPERTY_NAME);
            if (exchange != null) {
                final NormalizedMessage outNm = exchange.getMessage(JBI_OUT_MESSAGE);
                final NormalizedMessage inNm = exchange.getMessage(JBI_IN_MESSAGE);
                final NormalizedMessage faultNm = exchange.getFault();
                
                final Source source =
                        faultNm != null ? faultNm.getContent() :
                            outNm != null ? outNm.getContent() :
                                (inNm != null && exchange.getStatus().equals(ExchangeStatus.ACTIVE)) ? inNm.getContent() : null;
                                
                if (source != null) {
                    final TraceCode traceCode = (TraceCode) logData
                            .get(FlowLogData.FLOW_STEP_TRACE_CODE);

                    assert traceCode != null : "A MONIT log event MUST always have an attribute 'traceCode' set !!!";
    
                    final File dumpFile = this.getDumpFile(traceCode, flowInstanceId, flowStepId);
                    logData.putData(PAYLOAD_CONTENT_DUMP_FILE_LOGDATA_NAME,
                            FileSystemHelper.getRelativePath(this.basedir.getAbsoluteFile(), dumpFile));
                    dumpFile.getParentFile().mkdirs();
    
                    try {
                        Source faultMessageContentAsSource = helpers.fork(source);
                        helpers.toFile(faultMessageContentAsSource, dumpFile);
                    } catch (final Exception e) {
                        reportError(e.getMessage(), e, ErrorManager.WRITE_FAILURE);
                    }
                }
            }
        }
    }

    /**
     * <p>
     * Generate a {@link File} with an absolute path to dump message into the directory associated to the given flow
     * instance identifier.
     * </p>
     * 
     * @param providedFlowInstanceId
     *            A flow instance identifier. If <code>null</code>, the flow instance identifier is retrieved from the
     *            current context.
     * 
     * @return
     * @throws IOException
     */
    private final File getDumpFile(final TraceCode traceCode, final String providedFlowInstanceId,
            final String flowStepId) {

        final String flowInstanceId;
        if (providedFlowInstanceId == null) {
            final Properties contextualProperties = PropertiesHelper
                    .flattenProperties(ExecutionContext.getProperties());
            flowInstanceId = contextualProperties.getProperty(FLOW_INSTANCE_ID_PROPERTY_NAME);
        } else {
            flowInstanceId = providedFlowInstanceId;
        }
        final String dumpFileName = PetalsPayloadDumperFileHandler.getContextualDumpFileName(traceCode, flowStepId);
        return new File(new File(this.dumpBasedir, flowInstanceId), dumpFileName);
    }

    /**
     * Compute the dump file name:
     * <code>&lt;flow-step-id&gt;_&lt;trace-code&gt;.xml</code> where:
     * <ul>
     * <li><code>flow-step-id</code> is the flow step identifier, or
     * <code>unknown</code> if the flow step identifier exists,</li>
     * <li><code>trace-code</code> is a code defining the beginning or ending of
     * the current flow step.</li>
     * </ul>
     * 
     * @param traceCode
     *            A code defining the beginning or ending of the current flow
     *            step
     * @return The dump file name to save the content of the message
     */
    private static final String getContextualDumpFileName(final TraceCode traceCode,
            final String flowStepId) {
        return (flowStepId == null ? "unknown" : flowStepId) + FILENAME_SEPARATOR + traceCode
                + DUMP_FILE_EXTENSION;
    }
}
