/**
 * Copyright (c) 2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.log.handler;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.logging.Handler;

import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.sax.SAXSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.apache.commons.pool.BasePoolableObjectFactory;
import org.apache.commons.pool.ObjectPool;
import org.apache.commons.pool.impl.GenericObjectPool;
import org.xml.sax.InputSource;

import com.ebmwebsourcing.easycommons.stream.InputStreamForker;
import com.ebmwebsourcing.easycommons.stream.ReaderInputStream;

/**
 * 
 * Homemade version of easycommons-utils SourceHelper.
 * 
 * Needed because we don't want easycommons-utils and its transformer pool arriving in the system classloader.
 * 
 * This class should be used only by {@link PetalsPayloadDumperFileHandler}, so its visibility and methods' visibilities
 * are restricted to package scope.
 * 
 * @author vnoel
 *
 */
class SourceHelpers {

    private final ObjectPool<Transformer> transformers;

    /**
     * 
     * @param maxActive
     *            maximum number of instances
     * @param maxIdle
     *            maximum number of idle instances
     */
    public SourceHelpers(final int maxActive, final int maxIdle) {
        this.transformers = new GenericObjectPool<>(new TransformersFactory(), maxActive,
                GenericObjectPool.DEFAULT_WHEN_EXHAUSTED_ACTION, GenericObjectPool.DEFAULT_MAX_WAIT, maxIdle);
    }

    public void toFile(final Source source, final File fileToWrite) throws Exception {
        final Transformer transformer = transformers.borrowObject();
        try (final FileWriter writer = new FileWriter(fileToWrite)) {
            transformer.transform(source, new StreamResult(writer));
        } finally {
            transformers.returnObject(transformer);
        }
    }

    /**
     * TODO should I use the XMLReader?
     * 
     * @param source
     * @return
     * @throws IOException
     */
    private final StreamSource forkSAXSource(final SAXSource source) throws IOException {
        final InputSource inputSource = source.getInputSource();
        final InputStreamForker streamForker;
        if (inputSource.getCharacterStream() != null) {
            streamForker = new InputStreamForker(new ReaderInputStream(inputSource.getCharacterStream()));
            // let's replace the input stream with a fork
            source.setInputSource(new InputSource(streamForker.fork()));
        } else {
            streamForker = new InputStreamForker(inputSource.getByteStream());
            // let's replace the input stream with a fork
            source.setInputSource(new InputSource(new InputStreamReader(streamForker.fork())));
        }
        // and return another stream for the fork
        return new StreamSource(streamForker.fork());
    }

    private StreamSource forkStreamSource(final StreamSource source) throws IOException {
        final InputStreamForker streamForker;
        if (source.getInputStream() != null) {
            streamForker = new InputStreamForker(source.getInputStream());
            // let's replace the input stream with a fork
            source.setInputStream(streamForker.fork());
        } else {
            streamForker = new InputStreamForker(new ReaderInputStream(source.getReader()));
            // let's replace the input stream with a fork
            source.setReader(new InputStreamReader(streamForker.fork()));
        }
        // and return another stream for the fork
        return new StreamSource(streamForker.fork());
    }

    /**
     * Fork, if necessary, a {@link Source} so that consuming the forked one does not consume the original one.
     * 
     * @param source
     *            {@link Source} to be forked.
     * @return Forked {@link Source}.
     * @throws <IOException>
     */
    public Source fork(Source source) throws IOException {
        if (source instanceof DOMSource) {
            // DOMSource are in-memory
            return source;
        } else if (source instanceof StreamSource) {
            return forkStreamSource((StreamSource) source);
        } else if (source instanceof SAXSource) {
            return forkSAXSource((SAXSource) source);
        } else {
            // staxsources can't be forked as we need to read them to fork them but then we can't reset the staxsource
            // with one of the forks... TODO or we need to work directly at the message exchange level? or anyway
            // staxsource are not meant to be used in the context of something like petals!!! (small sized messages)
            throw new IllegalArgumentException(
                    "Only DOMSource, StreamSource and SAXSource are supported right now, got a "
                            + source.getClass().getName());
        }
    }

    private static class TransformersFactory extends BasePoolableObjectFactory<Transformer> {

        /**
         * this should be instantiated only once with the logging {@link Handler} that use {@link SourceHelpers} and
         * thus using its classloader to choose the {@link TransformerFactory}.
         */
        private final TransformerFactory factory = TransformerFactory.newInstance();

        @Override
        public Transformer makeObject() throws Exception {
            return factory.newTransformer();
        }

    }

}
